from eventsourcing.tests.unit_test_cases import BasicStoredEventRepositoryTestCase, \
    ConcurrentStoredEventRepositoryTestCase, SimpleStoredEventIteratorTestCase, ThreadedStoredEventIteratorTestCase
from eventsourcing.tests.unit_test_cases_cassandra import CassandraStoredEventRepoTestCase


class TestCassandraStoredEventRepository(CassandraStoredEventRepoTestCase, BasicStoredEventRepositoryTestCase):
    pass


class TestSimpleStoredEventIteratorWithCassandra(CassandraStoredEventRepoTestCase, SimpleStoredEventIteratorTestCase):
    pass


class TestThreadedStoredEventIteratorWithCassandra(CassandraStoredEventRepoTestCase, ThreadedStoredEventIteratorTestCase):
    pass


class TestConcurrentStoredEventRepositoryWithCassandra(CassandraStoredEventRepoTestCase, ConcurrentStoredEventRepositoryTestCase):
    pass
